@@ -1,5 +1,5 @@
 macro_rules! impl_vec_field {
-    ($($docs:literal)* $name:ident: $ty:tt => $docker_name:literal) => {
+    ($(#[doc = $docs:expr])* $name:ident: $ty:tt => $docker_name:literal) => {
         paste::item! {
             $(
                 #[doc= $docs]
@@ -17,7 +17,7 @@ macro_rules! impl_vec_field {
 }
 
 macro_rules! impl_field {
-    ($($docs:literal)* $name:ident: $ty:ty => $docker_name:literal) => {
+    ($(#[doc = $docs:expr])* $name:ident: $ty:ty => $docker_name:literal) => {
         paste::item! {
             $(
                 #[doc= $docs]
@@ -32,7 +32,7 @@ macro_rules! impl_field {
 }
 
 macro_rules! impl_str_field {
-    ($($docs:literal)* $name:ident: $ty:tt => $docker_name:literal) => {
+    ($(#[doc = $docs:expr])* $name:ident: $ty:tt => $docker_name:literal) => {
         paste::item! {
             $(
                 #[doc= $docs]
@@ -49,7 +49,7 @@ macro_rules! impl_str_field {
 }
 
 macro_rules! impl_url_str_field {
-    ($($docs:literal)* $name:ident: $ty:tt => $docker_name:literal) => {
+    ($(#[doc = $docs:expr])* $name:ident: $ty:tt => $docker_name:literal) => {
         paste::item! {
             $(
                 #[doc= $docs]
@@ -66,7 +66,7 @@ macro_rules! impl_url_str_field {
 }
 
 macro_rules! impl_url_field {
-    ($($docs:literal)* $name:ident : $ty:tt => $docker_name:literal) => {
+    ($(#[doc = $docs:expr])* $name:ident : $ty:tt => $docker_name:literal) => {
         paste::item! {
             $(
                 #[doc= $docs]
@@ -80,7 +80,7 @@ macro_rules! impl_url_field {
 }
 
 macro_rules! impl_url_bool_field {
-    ($($docs:literal)* $name:ident => $docker_name:literal) => {
+    ($(#[doc = $docs:expr])* $name:ident => $docker_name:literal) => {
         paste::item! {
             $(
                 #[doc= $docs]
@@ -95,7 +95,7 @@ macro_rules! impl_url_bool_field {
 
 #[allow(unused_macros)]
 macro_rules! impl_str_enum_field {
-    ($($docs:literal)* $name:ident: $ty:tt => $docker_name:literal) => {
+    ($(#[doc = $docs:expr])* $name:ident: $ty:tt => $docker_name:literal) => {
         paste::item! {
             $(
                 #[doc= $docs]
@@ -110,7 +110,13 @@ macro_rules! impl_str_enum_field {
 }
 
 macro_rules! impl_map_field {
-    ($($docs:literal)* $name:ident: $ty:tt => $docker_name:literal) => {
+    (url $(#[doc = $docs:expr])* $name:ident: $ty:tt => $docker_name:literal) => {
+        impl_map_field! { $(#[doc = $docs])* $name: $ty => $docker_name => serde_json::to_string(&$name.into_iter().collect::<std::collections::HashMap<_, _>>()).unwrap_or_default() }
+    };
+    (json $(#[doc = $docs:expr])* $name:ident: $ty:tt => $docker_name:literal) => {
+        impl_map_field! { $(#[doc = $docs])* $name: $ty => $docker_name => serde_json::json!($name.into_iter().collect::<std::collections::HashMap<_, _>>()) }
+    };
+    ($(#[doc = $docs:expr])* $name:ident: $ty:tt => $docker_name:literal => $ret:expr) => {
         paste::item! {
             $(
                 #[doc= $docs]
@@ -121,24 +127,23 @@ macro_rules! impl_map_field {
                 K: AsRef<str> + serde::Serialize + Eq + std::hash::Hash,
                 V: AsRef<str> + serde::Serialize
             {
-                self.params.insert($docker_name, serde_json::json!($name.into_iter().collect::<std::collections::HashMap<_, _>>()));
+                self.params.insert($docker_name, $ret);
                 self
             }
         }
     };
 }
 
-macro_rules! impl_json_opts_builder {
-    ($($docs:literal)* $name:ident) => {
+macro_rules! impl_opts_builder {
+    ($(#[doc = $docs:expr])* $name:ident $ty:expr) => {
         paste::item! {
             $(
                 #[doc= $docs]
             )*
-            #[derive(serde::Serialize, Debug)]
+            #[derive(serde::Serialize, Debug, Default)]
             pub struct [< $name Opts >] {
-                params: std::collections::HashMap<&'static str, serde_json::Value>,
+                params: std::collections::HashMap<&'static str, $ty>,
             }
-
             impl [< $name Opts >] {
                 calculated_doc!{
                 #[doc = concat!("Returns a new instance of a builder for ", stringify!($name), "Opts.")]
@@ -146,18 +151,13 @@ macro_rules! impl_json_opts_builder {
                     [< $name OptsBuilder >]::default()
                 }
                 }
-
-                /// Serialize options as a JSON String. Returns None if no options are defined.
-                pub fn serialize(&self) -> crate::Result<String> {
-                    serde_json::to_string(&self.params).map_err(crate::Error::from)
-                }
             }
 
             calculated_doc!{
             #[doc = concat!("A builder struct for ", stringify!($name), "Opts.")]
             #[derive(Default, Debug)]
             pub struct [< $name OptsBuilder >] {
-                params: std::collections::HashMap<&'static str, serde_json::Value>,
+                params: std::collections::HashMap<&'static str, $ty>,
             }
             }
 
@@ -171,32 +171,26 @@ macro_rules! impl_json_opts_builder {
                 }
                 }
             }
+       }
+    };
+    (json => $(#[doc = $docs:expr])* $name:ident) => {
+        paste::item! {
+            impl_opts_builder!($(#[doc = $docs])* $name serde_json::Value);
+
+            impl [< $name Opts >] {
+                /// Serialize options as a JSON String. Returns an error if the options will fail
+                /// to serialize.
+                pub fn serialize(&self) -> crate::Result<String> {
+                    serde_json::to_string(&self.params).map_err(crate::Error::from)
+                }
+            }
         }
     };
-}
-
-macro_rules! impl_url_opts_builder {
-    ($($docs:literal)* $name:ident) => {
-        impl_url_opts_builder!{derives = | $($docs)* $name}
-    };
-    (derives = $($derives:ident),* | $($docs:literal)* $name:ident) => {
+    (url => $(#[doc = $docs:expr])* $name:ident) => {
         paste::item! {
-            $(
-                #[doc= $docs]
-            )*
-            #[derive(serde::Serialize, Debug $(,$derives)*)]
-            pub struct [< $name Opts >] {
-                params: std::collections::HashMap<&'static str, String>
-            }
+            impl_opts_builder!($(#[doc = $docs])* $name String);
 
             impl [< $name  Opts >] {
-                calculated_doc!{
-                #[doc = concat!("Returns a new instance of a builder for ", stringify!($name), "Opts.")]
-                pub fn builder() -> [< $name  OptsBuilder >] {
-                  [< $name  OptsBuilder >]::default()
-                }
-                }
-
                 /// Serialize options as a URL query String. Returns None if no options are defined.
                 pub fn serialize(&self) -> Option<String> {
                     if self.params.is_empty() {
@@ -208,25 +202,6 @@ macro_rules! impl_url_opts_builder {
                     }
                 }
             }
-
-            calculated_doc!{
-            #[doc = concat!("A builder struct for ", stringify!($name), "Opts.")]
-            #[derive(Default, Debug)]
-            pub struct [< $name  OptsBuilder >] {
-                params: std::collections::HashMap<&'static str, String>
-            }
-            }
-
-            impl [< $name  OptsBuilder >] {
-                calculated_doc!{
-                #[doc = concat!("Finish building ", stringify!($name), "Opts.")]
-                pub fn build(&self) -> [< $name  Opts >] {
-                    [< $name Opts >] {
-                        params: self.params.clone(),
-                    }
-                }
-                }
-            }
         }
     };
 }
@@ -247,7 +222,7 @@ macro_rules! calculated_doc {
 }
 
 macro_rules! impl_api_ty {
-    ($($docs:literal)* $name:ident => $name_field:ident : $name_field_tt:tt) => {
+    ($(#[doc = $docs:expr])* $name:ident => $name_field:ident : $name_field_tt:tt) => {
         paste::item! {
             calculated_doc!{
             #[doc = concat!("Interface for accessing and manipulating Docker ", stringify!($name), ".\n", $($docs,)* "\n", api_url!($name))]
@@ -315,7 +290,10 @@ macro_rules! impl_api_ty {
 }
 
 macro_rules! impl_filter_func {
-    ($filter_ty:ident) => {
+    ($(#[doc = $doc:expr])* $filter_ty:ident) => {
+        $(
+            #[doc = $doc]
+        )*
         pub fn filter<F>(&mut self, filters: F) -> &mut Self
         where
             F: IntoIterator<Item = $filter_ty>,
@@ -403,9 +381,8 @@ macro_rules! impl_api_ep {
         #[doc = concat!("Inspect this ", stringify!($base), ".")]
         |
         pub async fn inspect(&self) -> Result<[< $base $ret >]> {
-            let ep_fn = Box::new(|$it: &$base| $ep);
-            let ep: String = ep_fn(self);
-            self.docker.get_json(ep.as_ref()).await
+            let $it = self;
+            self.docker.get_json($ep).await
         }}
         }
     };
@@ -426,9 +403,8 @@ macro_rules! impl_api_ep {
                 None
             };
 
-            let ep_fn= Box::new(|$it: &$base| $ep);
-            let ep = ep_fn(self);
-            let ep = crate::util::url::construct_ep(ep, query);
+            let $it = self;
+            let ep = crate::util::url::construct_ep($ep, query);
 
             self.docker
                 .delete_json(ep.as_ref())
@@ -460,14 +436,18 @@ macro_rules! impl_api_ep {
         #[doc = concat!("Delete this ", stringify!($base), ".")]
         |
         pub async fn delete(&self) -> Result<()> {
-            let ep_fn= Box::new(|$it: &$base| $ep);
-            let ep = ep_fn(self);
-            self.docker.delete(ep.as_ref()).await.map(|_| ())
+            let $it = self;
+            self.docker.delete($ep).await.map(|_| ())
         }}
         }
     };
     (
         DeleteWithOpts $it:ident: $base:ident -> $resp:ident $ep:expr, $ret:tt $(,$extra:expr)*
+    ) => {
+        impl_api_ep! { DeleteWithOpts $it: $base -> $resp $ep, $ret => $($extra)* }
+    };
+    (
+        DeleteWithOpts $it:ident: $base:ident -> $resp:ident $ep:expr, $ret:tt => $fn:expr
     ) => {
         paste::item! {
         api_doc! { $base => Delete
@@ -475,10 +455,9 @@ macro_rules! impl_api_ep {
         #[doc = concat!("Use [`delete`](", stringify!($base), "::delete) to delete without options.")]
         |
         pub async fn remove(&self, opts: &[< Rm $base Opts >]) -> Result<[< $ret >]> {
-            let ep_fn= Box::new(|$it: &$base| $ep);
-            let ep = ep_fn(self);
-            let ep = crate::util::url::construct_ep(ep, opts.serialize());
-            self.docker.delete_json(ep.as_ref()).await
+            let $it = self;
+            let ep = crate::util::url::construct_ep($ep, opts.serialize());
+            self.docker.$fn(ep.as_ref()).await
         }}
         }
         paste::item! {
@@ -487,9 +466,8 @@ macro_rules! impl_api_ep {
         #[doc = concat!("Use [`remove`](", stringify!($base), "::remove) to customize options.")]
         |
         pub async fn delete(&self) -> Result<[< $ret >]> {
-            let ep_fn= Box::new(|$it: &$base| $ep);
-            let ep = ep_fn(self);
-            self.docker.delete_json(ep.as_ref()).await
+            let $it = self;
+            self.docker.$fn($ep).await
         }}
         }
     };
@@ -509,7 +487,15 @@ macro_rules! impl_api_ep {
     (
         List $it:ident: $base:ident -> $resp:ident $ep:expr $(, $extra:expr)*
     ) => {
-        impl_api_ep! { List $it: $base -> $resp $ep, [< $base  Info >] }
+        paste::item! {
+        api_doc! { $base => List
+        #[doc = concat!("List available ", stringify!($base), "s.")]
+        |
+        pub async fn list(&self, opts: &[< $base ListOpts >]) -> Result<Vec<[< $base Info >]>> {
+            let ep = crate::util::url::construct_ep($ep, opts.serialize());
+            self.docker.get_json(&ep).await
+        }}
+        }
     };
     (
         Create $it:ident: $base:ident -> $resp:ident $ep:expr $(, $extra:expr)*
@@ -551,9 +537,8 @@ macro_rules! impl_api_ep {
             &self,
             opts: &crate::api::LogsOpts
         ) -> impl futures_util::Stream<Item = crate::Result<crate::conn::TtyChunk>> + Unpin + 'docker {
-            let ep_fn= Box::new(|$it: &[< $base >]| $ep);
-            let ep = ep_fn(self);
-            let ep = crate::util::url::construct_ep(ep, opts.serialize());
+            let $it = self;
+            let ep = crate::util::url::construct_ep($ep, opts.serialize());
 
             let stream = Box::pin(self.docker.stream_get(ep));
 
